unit FastBlend; // FastBlend
                // Updated: 5/11/2001
interface       // 2001 g-soft http://gfody.com

{$R-}

uses Windows, FastDIB;

procedure AvgBlend(Dst,Src1,Src2:TFastDIB);
procedure DifBlend(Dst,Src1,Src2:TFastDIB);
procedure SubBlend(Dst,Src1,Src2:TFastDIB);
procedure AddBlend(Dst,Src1,Src2:TFastDIB);
procedure MulBlend(Dst,Src1,Src2:TFastDIB);
procedure AlphaBlend(Dst,Src1,Src2:TFastDIB;Alpha:Integer);

implementation

procedure AvgMem(Dst,Src1,Src2:Pointer;Size,Mask:Integer);
// Dst  = eax
// Src1 = edx
// Src2 = ecx
// Size = [ebp + 12]
// Mask = [ebp +  8]
// s    = [ebp -  4]
// a    = [ebp -  8]
// b    = [ebp - 12]
var
  s,a,b: Integer;
asm
  push ebx
  push edi
  push esi

  mov edi,edx
  mov esi,ecx
  mov ecx,[ebp+12]
  push ecx
  mov s,esp
  mov esp,eax
  and ecx,$FFFFFFFC
  jz  @skip

  add ecx,eax
  mov [ebp+12],ecx
  mov eax,[ebp+8]

  // (A+B)/2 = (A and B)+((A xor B)/2) : Paul Hsieh
  @dwords:
    mov ebx,[edi]
    mov edx,[esi]
    mov ecx,ebx
    and ebx,edx
    xor edx,ecx
    shr edx,1
    and edx,eax
    add edx,ebx
    mov [esp],edx
    add edi,4
    add esi,4
    add esp,4
    cmp esp,[ebp+12]
  jne @dwords

  @skip:
  mov eax,esp
  mov esp,s
  pop ecx
  and ecx,11b
  jz  @exit

  mov edx,ecx
  push edi
  lea edi,a
  rep movsb

  mov ecx,edx
  pop esi
  lea edi,b
  rep movsb

  mov ebx,b
  mov ecx,a
  and ebx,ecx
  xor ecx,b
  shr ecx,1
  and ecx,[ebp+8]
  add ecx,ebx
  mov b,ecx

  mov ecx,edx
  lea esi,b
  mov edi,eax
  rep movsb

  @exit:
  pop esi
  pop edi
  pop ebx
end;

procedure AvgMemMMX(Dst,Src1,Src2:Pointer;Size,Mask:Integer);
// Dst  = eax
// Src1 = edx
// Src2 = ecx
// Size = [ebp + 12]
// Mask = [ebp +  8]
// a:b  = [ebp -  8]
var
  a,b: Integer;
asm
  push edi
  push esi

  mov a,0
  mov b,0

  mov edi,edx
  mov esi,ecx
  mov ecx,[ebp+12]

  db $0F,$6E,$5D,$08 /// movd  mm3,[ebp+8]
  db $0F,$6F,$E3     /// movq  mm4,mm3
  db $0F,$73,$F4,$20 /// psllq mm4,32
  db $0F,$EB,$DC     /// por   mm3,mm4

  shr ecx,3
  jz  @skip

  @quads:
    db $0F,$6F,$07     /// movq  mm0,[edi]
    db $0F,$6F,$0E     /// movq  mm1,[esi]
    db $0F,$6F,$D0     /// movq  mm2,mm0
    db $0F,$DB,$C1     /// pand  mm0,mm1
    db $0F,$EF,$CA     /// pxor  mm1,mm2
    db $0F,$73,$D1,$01 /// psrlq mm1,1
    db $0F,$DB,$CB     /// pand  mm1,mm3
    db $0F,$FE,$C8     /// paddd mm1,mm0
    db $0F,$7F,$08     /// movq  [eax],mm1
    add edi,8
    add esi,8
    add eax,8
    dec ecx
  jnz @quads

  @skip:
  mov ecx,[ebp+12]
  and ecx,111b
  jz  @exit

  mov edx,ecx
  push edi
  lea edi,[ebp-8]
  rep movsb
  db $0F,$6F,$8D,$F8,$FF,$FF,$FF /// movq mm1,[ebp-8]

  mov ecx,edx
  pop esi
  lea edi,[ebp-8]
  rep movsb
  db $0F,$6F,$85,$F8,$FF,$FF,$FF /// movq mm0,[ebp-8]

  db $0F,$6F,$D0                 /// movq  mm2,mm0
  db $0F,$DB,$C1                 /// pand  mm0,mm1
  db $0F,$EF,$CA                 /// pxor  mm1,mm2
  db $0F,$73,$D1,$01             /// psrlq mm1,1
  db $0F,$DB,$CB                 /// pand  mm1,mm3
  db $0F,$FE,$C8                 /// paddd mm1,mm0
  db $0F,$7F,$8D,$F8,$FF,$FF,$FF /// movq  [ebp-8],mm1

  mov ecx,edx
  lea esi,[ebp-8]
  mov edi,eax
  rep movsb

  @exit:
  db $0F,$77 // emms
  pop esi
  pop edi
end;

procedure AvgBlend(Dst,Src1,Src2:TFastDIB);
type
  TAvgMemProc = procedure(Dst,Src1,Src2:Pointer;Size,Mask:Integer);
var
  i,Mask: Integer;
  AvgProc: TAvgMemProc;
begin
  case Dst.Bpp of
    8,24: Mask:=$7F7F7F7F;
    16,32:
    begin
      Mask:=((Dst.RMask+(1 shl Dst.RShl))or(Dst.GMask+(1 shl Dst.GShl))or(Dst.BMask+1))shr 1;
      if Dst.Bpp=16 then Mask:=(Mask shl 16 or Mask)xor -1 else Mask:=Mask xor -1;
    end;
    else Mask:=0;
  end;

  if cfMMX in CPUInfo.Features then
    AvgProc:=AvgMemMMX else AvgProc:=AvgMem;

  for i:=0 to Dst.AbsHeight-1 do
    AvgProc(Dst.Scanlines[i],Src1.Scanlines[i],Src2.Scanlines[i],Dst.BWidth-Dst.Gap,Mask);
end;

{
procedure DifMem16(Dst,Src1,Src2:Pointer;Size:Integer);
// Dst  = eax
// Src1 = edx
// Src2 = ecx
// Size = [ebp +  8]
// a    = [ebp -  4]
// b    = [ebp -  8]
// c    = [ebp - 12]
var
  a,b,c: Integer;
asm
  push ebx
  push edi
  push esi

  mov  edi,edx
  mov  esi,ecx
  mov  ecx,[ebp+8]
  push ecx
  and  ecx,$FFFFFFFC
  jz   @skip

  add ecx,eax
  mov [ebp+8],ecx

  @dwords:
    mov ecx,[edi]
    mov ebx,[esi]
    and ecx,$001F001F
    and ebx,$001F001F
    or  ecx,$00200020
    sub ecx,ebx
    mov ebx,ecx
    and ebx,$00200020
    shr ebx,5
    imul ebx,$1F
    xor ecx,ebx

    mov edx,[edi]
    mov ebx,[esi]
    and edx,$07E007E0
    and ebx,$07E007E0
    shr edx,5
    shr ebx,5
    or  edx,$00400040
    sub edx,ebx
    mov ebx,edx
    and ebx,$00400040
    shr ebx,6
    imul ebx,$3F
    xor edx,ebx
    shl edx,5
    or  ecx,edx

    mov edx,[edi]
    mov ebx,[esi]
    and edx,$F800F800
    and ebx,$F800F800
    shr edx,11
    shr ebx,11
    or  edx,$00200020
    sub edx,ebx
    mov ebx,edx
    and ebx,$00200020
    shr ebx,5
    imul ebx,$1F
    xor edx,ebx
    shl edx,11
    or  ecx,edx
    xor ecx,-1

    mov [eax],ecx
    add edi,4
    add esi,4
    add eax,4
    cmp eax,[ebp+8]
  jne @dwords

  cmp edi,ebp
  je  @last

  @skip:
  pop ecx
  and ecx,11b
  shr ecx,1
  jz  @exit

  mov  cx,[edi]
  mov  bx,[esi]
  mov  a,ecx
  mov  b,ebx
  lea  edi,a
  lea  esi,b
  push eax
  lea  eax,c
  lea  ebx,[eax+4]
  mov  [ebp+8],ebx
  jmp  @dwords

  @last:
  pop eax
  mov ebx,[ebp-12]
  mov [eax],bx

  @exit:
  pop esi
  pop edi
  pop ebx
end;

procedure DifMem16MMX(Dst,Src1,Src2:Pointer;Size:Integer);
// Dst   = eax
// Src1  = edx
// Src2  = ecx
// Size  = [ebp +  8]
// B1:B2 = [ebp -  8]
// G1:G2 = [ebp - 16]
// M1:M2 = [ebp - 24]
var
  B1,B2,G1,G2,M1,M2: Integer;
asm
  push ebx
  push edi
  push esi

  mov B1,$001F001F
  mov B2,$001F001F
  mov G1,$07E007E0
  mov G2,$07E007E0
  mov M2,0
  mov M1,0

  pcmpeqd mm7,mm7
  psrlw   mm7,11
  psllw   mm7,11

  mov edi,edx
  mov esi,ecx
  mov ecx,[ebp+8]
  push ecx
  shr ecx,3
  jz  @skip

  @quads:
    movq    mm0,[edi]
    movq    mm1,[esi]
  @cleanup:
    movq    mm2,mm0
    movq    mm3,mm1
    movq    mm4,mm0
    movq    mm5,mm1

    pand    mm0,[ebp-8]
    pand    mm1,[ebp-8]
    movq    mm6,mm0
    psubusw mm0,mm1
    psubusw mm1,mm6
    paddusw mm0,mm1

    pand    mm2,[ebp-16]
    pand    mm3,[ebp-16]
    psrlw   mm2,5
    psrlw   mm3,5
    movq    mm6,mm2
    psubusw mm2,mm3
    psubusw mm3,mm6
    paddusw mm2,mm3
    psllw   mm2,5
    por     mm0,mm2

    pand    mm4,mm7
    pand    mm5,mm7
    psrlw   mm4,11
    psrlw   mm5,11
    movq    mm6,mm4
    psubusw mm4,mm5
    psubusw mm5,mm6
    paddusw mm4,mm5
    psllw   mm4,11
    por     mm0,mm4

    movq   [eax],mm0

    add edi,8
    add esi,8
    add eax,8
    dec ecx
  jnz @quads

  cmp edi,0
  je @last

  @skip:
  pop ecx
  and ecx,111b
  shr ecx,1
  jz  @exit

  mov  edx,ecx
  push edi
  lea  edi,M1
  rep  movsw
  movq mm1,[ebp-24]
  pop  esi
  mov  ecx,edx
  lea  edi,M1
  rep  movsw
  movq mm0,[ebp-24]
  push eax
  lea  eax,M1
  mov  edi,-8
  mov  ecx,1
  jmp  @cleanup

  @last:
  pop edi
  lea esi,M1
  mov ecx,edx
  rep movsw

  @exit:
  emms
  pop esi
  pop edi
  pop ebx
end;
}

procedure DifBlend16(Dst,Src1,Src2:TFastDIB);
type
  TDifMem16REG = array[0..256]of Byte; PDifMem16REG =^TDifMem16REG;
  TDifMem16MMX = array[0..299]of Byte; PDifMem16MMX =^TDifMem16MMX;
  TDifMem16Proc = procedure(Dst,Src1,Src2:Pointer;Size:Integer);
const
  DifMem16REG: TDifMem16REG =
  ($55,$8B,$EC,$83,$C4,$F4,$53,$57,$56,$89,$D7,$89,$CE,$8B,$4D,$08,$51,$83,$E1,
   $FC,$0F,$84,$AE,$00,$00,$00,$01,$C1,$89,$4D,$08,$8B,$0F,$8B,$1E,$81,$E1,$1F,
   $00,$1F,$00,$81,$E3,$1F,$00,$1F,$00,$81,$C9,$20,$00,$20,$00,$29,$D9,$89,$CB,
   $81,$E3,$20,$00,$20,$00,$C1,$EB,$05,$6B,$DB,$1F,$31,$D9,$8B,$17,$8B,$1E,$81,
   $E2,$E0,$07,$E0,$07,$81,$E3,$E0,$07,$E0,$07,$C1,$EA,$05,$C1,$EB,$05,$81,$CA,
   $40,$00,$40,$00,$29,$DA,$89,$D3,$81,$E3,$40,$00,$40,$00,$C1,$EB,$06,$6B,$DB,
   $3F,$31,$DA,$C1,$E2,$05,$09,$D1,$8B,$17,$8B,$1E,$81,$E2,$00,$F8,$00,$F8,$81,
   $E3,$00,$F8,$00,$F8,$C1,$EA,$0B,$C1,$EB,$0B,$81,$CA,$20,$00,$20,$00,$29,$DA,
   $89,$D3,$81,$E3,$20,$00,$20,$00,$C1,$EB,$05,$6B,$DB,$1F,$31,$DA,$C1,$E2,$0B,
   $09,$D1,$83,$F1,$FF,$89,$08,$83,$C7,$04,$83,$C6,$04,$83,$C0,$04,$3B,$45,$08,
   $0F,$85,$5B,$FF,$FF,$FF,$39,$EF,$74,$29,$59,$83,$E1,$03,$D1,$E9,$74,$28,$66,
   $8B,$0F,$66,$8B,$1E,$89,$4D,$FC,$89,$5D,$F8,$8D,$7D,$FC,$8D,$75,$F8,$50,$8D,
   $45,$F4,$8D,$58,$04,$89,$5D,$08,$E9,$2E,$FF,$FF,$FF,$58,$8B,$5D,$F4,$66,$89,
   $18,$5E,$5F,$5B,$8B,$E5,$5D,$C2,$04,$00);

  DifMem16MMX: TDifMem16MMX =
  ($55,$8B,$EC,$83,$C4,$E8,$53,$57,$56,$C7,$45,$FC,$1F,$00,$1F,$00,$C7,$45,$F8,
   $1F,$00,$1F,$00,$C7,$45,$F4,$E0,$07,$E0,$07,$C7,$45,$F0,$E0,$07,$E0,$07,$C7,
   $45,$EC,$00,$00,$00,$00,$C7,$45,$E8,$00,$00,$00,$00,$0F,$76,$FF,$0F,$71,$D7,
   $0B,$0F,$71,$F7,$0B,$89,$D7,$89,$CE,$8B,$4D,$08,$51,$C1,$E9,$03,$0F,$84,$8E,
   $00,$00,$00,$0F,$6F,$07,$0F,$6F,$0E,$0F,$6F,$D0,$0F,$6F,$D9,$0F,$6F,$E0,$0F,
   $6F,$E9,$0F,$DB,$85,$F8,$FF,$FF,$FF,$0F,$DB,$8D,$F8,$FF,$FF,$FF,$0F,$6F,$F0,
   $0F,$D9,$C1,$0F,$D9,$CE,$0F,$DD,$C1,$0F,$DB,$95,$F0,$FF,$FF,$FF,$0F,$DB,$9D,
   $F0,$FF,$FF,$FF,$0F,$71,$D2,$05,$0F,$71,$D3,$05,$0F,$6F,$F2,$0F,$D9,$D3,$0F,
   $D9,$DE,$0F,$DD,$D3,$0F,$71,$F2,$05,$0F,$EB,$C2,$0F,$DB,$E7,$0F,$DB,$EF,$0F,
   $71,$D4,$0B,$0F,$71,$D5,$0B,$0F,$6F,$F4,$0F,$D9,$E5,$0F,$D9,$EE,$0F,$DD,$E5,
   $0F,$71,$F4,$0B,$0F,$EB,$C4,$0F,$7F,$00,$83,$C7,$08,$83,$C6,$08,$83,$C0,$08,
   $49,$0F,$85,$77,$FF,$FF,$FF,$83,$FF,$00,$74,$3B,$59,$83,$E1,$07,$D1,$E9,$74,
   $3C,$89,$CA,$57,$8D,$7D,$E8,$66,$F3,$A5,$0F,$6F,$8D,$E8,$FF,$FF,$FF,$5E,$89,
   $D1,$8D,$7D,$E8,$66,$F3,$A5,$0F,$6F,$85,$E8,$FF,$FF,$FF,$50,$8D,$45,$E8,$BF,
   $F8,$FF,$FF,$FF,$B9,$01,$00,$00,$00,$E9,$3D,$FF,$FF,$FF,$5F,$8D,$75,$E8,$89,
   $D1,$66,$F3,$A5,$0F,$77,$5E,$5F,$5B,$8B,$E5,$5D,$C2,$04,$00);
var
  i: Integer;
  Code: PLine8;
begin
  if cfMMX in CPUInfo.Features then
  begin
    GetMem(Code,SizeOf(TDifMem16MMX));
    PDifMem16MMX(Code)^:=DifMem16MMX;

    i:=Dst.BMask shl 16 or Dst.BMask;
    PDWord(@Code[12])^:=i;
    PDWord(@Code[19])^:=i;
    i:=Dst.GMask shl 16 or Dst.GMask;
    PDWord(@Code[26])^:=i;
    PDWord(@Code[33])^:=i;
    Code[57]:=16-Dst.Bpr;
    Code[61]:=Dst.RShl;
    Code[140]:=Dst.GShl;
    Code[144]:=Dst.GShl;
    Code[160]:=Dst.GShl;
    Code[173]:=Dst.RShl;
    Code[177]:=Dst.RShl;
    Code[193]:=Dst.RShl;
  end else
  begin
    GetMem(Code,SizeOf(TDifMem16REG));
    PDifMem16REG(Code)^:=DifMem16REG;

    i:=Dst.BMask shl 16 or Dst.BMask;
    PDWord(@Code[37])^:=i;
    PDWord(@Code[43])^:=i;
    i:=(i+i)and(not i);
    PDWord(@Code[49])^:=i;
    PDWord(@Code[59])^:=i;
    i:=Dst.GMask shl 16 or Dst.GMask;
    PDWord(@Code[77])^:=i;
    PDWord(@Code[83])^:=i;
    i:=i shr Dst.GShl;
    i:=(i+i)and(not i);
    PDWord(@Code[95])^:=i;
    PDWord(@Code[105])^:=i;
    i:=Dst.RMask shl 16 or Dst.RMask;
    PDWord(@Code[128])^:=i;
    PDWord(@Code[134])^:=i;
    i:=i shr Dst.RShl;
    i:=(i+i)and(not i);
    PDWord(@Code[146])^:=i;
    PDWord(@Code[156])^:=i;
    Code[65]:=Dst.Bpb;
    Code[68]:=(1 shl Dst.Bpb)-1;
    Code[89]:=Dst.GShl;
    Code[92]:=Dst.GShl;
    Code[111]:=Dst.Bpg;
    Code[114]:=(1 shl Dst.Bpg)-1;
    Code[119]:=Dst.GShl;
    Code[140]:=Dst.RShl;
    Code[143]:=Dst.RShl;
    Code[162]:=Dst.Bpr;
    Code[165]:=(1 shl Dst.Bpr)-1;
    Code[170]:=Dst.RShl;
  end;

  for i:=0 to Dst.AbsHeight-1 do
    TDifMem16Proc(Code)(Dst.Scanlines[i],Src1.Scanlines[i],Src2.Scanlines[i],Dst.BWidth-Dst.Gap);

  FreeMem(Code);
end;

procedure DifMem8(Dst,Src1,Src2:Pointer;Size:Integer);
// Dst  = eax
// Src1 = edx
// Src2 = ecx
// Size = [ebp + 8]
// s    = [ebp - 4]
var
  s: Integer;
asm
  push ebx
  push edi
  push esi

  mov edi,edx
  mov esi,ecx
  mov ecx,[ebp+8]
  push ecx
  mov s,esp
  mov esp,eax
  and ecx,$FFFFFFFC
  jz  @skip
  add ecx,eax
  mov [ebp+8],ecx
  
  @dwords:
    mov  ecx,[edi]
    mov  ebx,[esi]
    mov  eax,ecx
    mov  edx,ebx

    and  ecx,$00FF00FF
    and  ebx,$00FF00FF
    or   ecx,$01000100
    sub  ecx,ebx
    mov  ebx,ecx
    and  ebx,$01000100
    imul ebx,$FF
    shr  ebx,8
    xor  ecx,ebx

    and  eax,$FF00FF00
    and  edx,$FF00FF00
    shr  eax,8
    shr  edx,8
    or   eax,$01000100
    sub  eax,edx
    mov  edx,eax
    and  edx,$01000100
    imul edx,$FF
    shr  edx,8
    xor  eax,edx
    shl  eax,8
    or   ecx,eax
    xor  ecx,-1

    mov [esp],ecx
    add edi,4
    add esi,4
    add esp,4
    cmp esp,[ebp+8]
  jne @dwords

  @skip:
  mov eax,esp
  mov esp,s
  pop ecx
  and ecx,11b
  jz  @exit

  @bytes:
    movzx ebx,Byte([edi])
    movzx edx,Byte([esi])
    sub ebx,edx
    mov edx,ebx
    shr edx,8
    xor ebx,edx
    mov [eax],bl
    inc edi
    inc esi
    inc eax
    dec ecx
  jnz @bytes

  @exit:
  pop esi
  pop edi
  pop ebx
end;

procedure DifMem8MMX(Dst,Src1,Src2:Pointer;Size:Integer);
// Dst  = eax
// Src1 = edx
// Src2 = ecx
// Size = [ebp + 8]
asm
  push ebx
  push edi
  push esi

  mov edi,edx
  mov esi,ecx
  mov ecx,[ebp+8]
  push ecx
  shr ecx,3
  jz  @skip

  @quads:
    db $0F,$6F,$07  /// movq    mm0,[edi]
    db $0F,$6F,$0E  /// movq    mm1,[esi]
    db $0F,$6F,$D0  /// movq    mm2,mm0
    db $0F,$D8,$C1  /// psubusb mm0,mm1
    db $0F,$D8,$CA  /// psubusb mm1,mm2
    db $0F,$DC,$C1  /// paddusb mm0,mm1
    db $0F,$7F,$00  /// movq    [eax],mm0
    add edi,8
    add esi,8
    add eax,8
    dec ecx
  jnz @quads

  db $0F,$77 // emms

  @skip:
  pop ecx
  and ecx,111b
  jz  @exit

  @bytes:
    movzx ebx,Byte([edi])
    movzx edx,Byte([esi])
    sub ebx,edx
    mov edx,ebx
    shr edx,8
    xor ebx,edx
    mov [eax],bl
    inc edi
    inc esi
    inc eax
    dec ecx
  jnz @bytes

  @exit:
  pop esi
  pop edi
  pop ebx
end;

procedure DifBlend8(Dst,Src1,Src2:TFastDIB);
type
  TDifMem8Proc = procedure(Dst,Src1,Src2:Pointer;Size:Integer);
var
  i: Integer;
  DifMem: TDifMem8Proc;
begin
  if cfMMX in CPUInfo.Features then
    DifMem:=DifMem8MMX else DifMem:=DifMem8;

  for i:=0 to Dst.AbsHeight-1 do
    DifMem(Dst.Scanlines[i],Src1.Scanlines[i],Src2.Scanlines[i],Dst.BWidth-Dst.Gap);
end;

procedure DifBlend(Dst,Src1,Src2:TFastDIB);
begin
  case Dst.Bpp of
    8,24,32: DifBlend8(Dst,Src1,Src2);
    16: DifBlend16(Dst,Src1,Src2);
  end;
end;

{
procedure SubMem16(Dst,Src1,Src2:Pointer;Size:Integer);
// Dst  = eax
// Src1 = edx
// Src2 = ecx
// Size = [ebp +  8]
// a    = [ebp -  4]
// b    = [ebp -  8]
// c    = [ebp - 12]
var
  a,b,c: Integer;
asm
  push ebx
  push edi
  push esi

  mov  edi,edx
  mov  esi,ecx
  mov  ecx,[ebp+8]
  push ecx
  and  ecx,$FFFFFFFC
  jz   @skip

  add ecx,eax
  mov [ebp+8],ecx

  @dwords:
    mov ecx,[edi]
    mov ebx,[esi]
    and ecx,$001F001F
    and ebx,$001F001F
    or  ecx,$00200020
    sub ecx,ebx
    mov ebx,ecx
    and ebx,$00200020
    shr ebx,5
    imul ebx,$1F
    and ecx,ebx

    mov edx,[edi]
    mov ebx,[esi]
    and edx,$07E007E0
    and ebx,$07E007E0
    shr edx,5
    shr ebx,5
    or  edx,$00400040
    sub edx,ebx
    mov ebx,edx
    and ebx,$00400040
    shr ebx,6
    imul ebx,$3F
    and edx,ebx
    shl edx,5
    or  ecx,edx

    mov edx,[edi]
    mov ebx,[esi]
    and edx,$F800F800
    and ebx,$F800F800
    shr edx,11
    shr ebx,11
    or  edx,$00200020
    sub edx,ebx
    mov ebx,edx
    and ebx,$00200020
    shr ebx,5
    imul ebx,$1F
    and edx,ebx
    shl edx,11
    or  ecx,edx

    mov [eax],ecx
    add edi,4
    add esi,4
    add eax,4
    cmp eax,[ebp+8]
  jne @dwords

  cmp edi,ebp
  je  @last

  @skip:
  pop ecx
  and ecx,11b
  shr ecx,1
  jz  @exit

  mov  cx,[edi]
  mov  bx,[esi]
  mov  a,ecx
  mov  b,ebx
  lea  edi,a
  lea  esi,b
  push eax
  lea  eax,c
  lea  ebx,[eax+4]
  mov  [ebp+8],ebx
  jmp  @dwords

  @last:
  pop eax
  mov ebx,[ebp-12]
  mov [eax],bx

  @exit:
  pop esi
  pop edi
  pop ebx
end;

procedure SubMem16MMX(Dst,Src1,Src2:Pointer;Size:Integer);
// Dst   = eax
// Src1  = edx
// Src2  = ecx
// Size  = [ebp +  8]
// B1:B2 = [ebp -  8]
// M1:M2 = [ebp - 16]
var
  B1,B2,M1,M2: Integer;
asm
  push ebx
  push edi
  push esi

  mov B1,$001F001F
  mov B2,$001F001F
  mov M2,0
  mov M1,0

  pcmpeqd mm6,mm6
  psrlw   mm6,10
  psllw   mm6,5
  pcmpeqd mm7,mm7
  psrlw   mm7,11
  psllw   mm7,11

  mov edi,edx
  mov esi,ecx
  mov ecx,[ebp+8]
  push ecx
  shr ecx,3
  jz  @skip

  @quads:
    movq    mm0,[edi]
    movq    mm1,[esi]
  @cleanup:
    movq    mm2,mm0
    movq    mm3,mm1
    movq    mm4,mm0
    movq    mm5,mm1

    pand    mm0,[ebp-8]
    pand    mm1,[ebp-8]
    psubusw mm0,mm1

    pand    mm2,mm6
    pand    mm3,mm6
    psrlw   mm2,5
    psrlw   mm3,5
    psubusw mm2,mm3
    psllw   mm2,5
    por     mm0,mm2

    pand    mm4,mm7
    pand    mm5,mm7
    psrlw   mm4,11
    psrlw   mm5,11
    psubusw mm4,mm5
    psllw   mm4,11
    por     mm0,mm4

    movq [eax],mm0

    add edi,8
    add esi,8
    add eax,8
    dec ecx
  jnz @quads

  cmp edi,0
  je @last

  @skip:
  pop ecx
  and ecx,111b
  shr ecx,1
  jz  @exit

  mov  edx,ecx
  push edi
  lea  edi,M1
  rep  movsw
  movq mm1,[ebp-16]
  pop  esi
  mov  ecx,edx
  lea  edi,M1
  rep  movsw
  movq mm0,[ebp-16]
  push eax
  lea  eax,M1
  mov  edi,-8
  mov  ecx,1
  jmp  @cleanup

  @last:
  pop edi
  lea esi,M1
  mov ecx,edx
  rep movsw

  @exit:
  emms
  pop esi
  pop edi
  pop ebx
end;
}

procedure SubBlend16(Dst,Src1,Src2:TFastDIB);
type
  TSubMem16REG = array[0..253]of Byte; PSubMem16REG =^TSubMem16REG;
  TSubMem16MMX = array[0..253]of Byte; PSubMem16MMX =^TSubMem16MMX;
  TSubMem16Proc = procedure(Dst,Src1,Src2:Pointer;Size:Integer);
const
  SubMem16REG: TSubMem16REG =
  ($55,$8B,$EC,$83,$C4,$F4,$53,$57,$56,$89,$D7,$89,$CE,$8B,$4D,$08,$51,$83,$E1,
   $FC,$0F,$84,$AB,$00,$00,$00,$01,$C1,$89,$4D,$08,$8B,$0F,$8B,$1E,$81,$E1,$1F,
   $00,$1F,$00,$81,$E3,$1F,$00,$1F,$00,$81,$C9,$20,$00,$20,$00,$29,$D9,$89,$CB,
   $81,$E3,$20,$00,$20,$00,$C1,$EB,$05,$6B,$DB,$1F,$21,$D9,$8B,$17,$8B,$1E,$81,
   $E2,$E0,$07,$E0,$07,$81,$E3,$E0,$07,$E0,$07,$C1,$EA,$05,$C1,$EB,$05,$81,$CA,
   $40,$00,$40,$00,$29,$DA,$89,$D3,$81,$E3,$40,$00,$40,$00,$C1,$EB,$06,$6B,$DB,
   $3F,$21,$DA,$C1,$E2,$05,$09,$D1,$8B,$17,$8B,$1E,$81,$E2,$00,$F8,$00,$F8,$81,
   $E3,$00,$F8,$00,$F8,$C1,$EA,$0B,$C1,$EB,$0B,$81,$CA,$20,$00,$20,$00,$29,$DA,
   $89,$D3,$81,$E3,$20,$00,$20,$00,$C1,$EB,$05,$6B,$DB,$1F,$21,$DA,$C1,$E2,$0B,
   $09,$D1,$89,$08,$83,$C7,$04,$83,$C6,$04,$83,$C0,$04,$3B,$45,$08,$0F,$85,$5E,
   $FF,$FF,$FF,$39,$EF,$74,$29,$59,$83,$E1,$03,$D1,$E9,$74,$28,$66,$8B,$0F,$66,
   $8B,$1E,$89,$4D,$FC,$89,$5D,$F8,$8D,$7D,$FC,$8D,$75,$F8,$50,$8D,$45,$F4,$8D,
   $58,$04,$89,$5D,$08,$E9,$31,$FF,$FF,$FF,$58,$8B,$5D,$F4,$66,$89,$18,$5E,$5F,
   $5B,$8B,$E5,$5D,$C2,$04,$00);

  SubMem16MMX: TSubMem16MMX =
  ($55,$8B,$EC,$83,$C4,$F0,$53,$57,$56,$C7,$45,$FC,$1F,$00,$1F,$00,$C7,$45,$F8,
   $1F,$00,$1F,$00,$C7,$45,$F4,$00,$00,$00,$00,$C7,$45,$F0,$00,$00,$00,$00,$0F,
   $76,$F6,$0F,$71,$D6,$0A,$0F,$71,$F6,$05,$0F,$76,$FF,$0F,$71,$D7,$0B,$0F,$71,
   $F7,$0B,$89,$D7,$89,$CE,$8B,$4D,$08,$51,$C1,$E9,$03,$74,$67,$0F,$6F,$07,$0F,
   $6F,$0E,$0F,$6F,$D0,$0F,$6F,$D9,$0F,$6F,$E0,$0F,$6F,$E9,$0F,$DB,$85,$F8,$FF,
   $FF,$FF,$0F,$DB,$8D,$F8,$FF,$FF,$FF,$0F,$D9,$C1,$0F,$DB,$D6,$0F,$DB,$DE,$0F,
   $71,$D2,$05,$0F,$71,$D3,$05,$0F,$D9,$D3,$0F,$71,$F2,$05,$0F,$EB,$C2,$0F,$DB,
   $E7,$0F,$DB,$EF,$0F,$71,$D4,$0B,$0F,$71,$D5,$0B,$0F,$D9,$E5,$0F,$71,$F4,$0B,
   $0F,$EB,$C4,$0F,$7F,$00,$83,$C7,$08,$83,$C6,$08,$83,$C0,$08,$49,$75,$9E,$83,
   $FF,$00,$74,$3B,$59,$83,$E1,$07,$D1,$E9,$74,$3C,$89,$CA,$57,$8D,$7D,$F0,$66,
   $F3,$A5,$0F,$6F,$8D,$F0,$FF,$FF,$FF,$5E,$89,$D1,$8D,$7D,$F0,$66,$F3,$A5,$0F,
   $6F,$85,$F0,$FF,$FF,$FF,$50,$8D,$45,$F0,$BF,$F8,$FF,$FF,$FF,$B9,$01,$00,$00,
   $00,$E9,$64,$FF,$FF,$FF,$5F,$8D,$75,$F0,$89,$D1,$66,$F3,$A5,$0F,$77,$5E,$5F,
   $5B,$8B,$E5,$5D,$C2,$04,$00);
var
  i: Integer;
  Code: PLine8;
begin
  if cfMMX in CPUInfo.Features then
  begin
    GetMem(Code,SizeOf(TSubMem16MMX));
    PSubMem16MMX(Code)^:=SubMem16MMX;

    i:=Dst.BMask shl 16 or Dst.BMask;
    PDWord(@Code[12])^:=i;
    PDWord(@Code[19])^:=i;
    Code[43]:=16-Dst.Bpg;
    Code[47]:=Dst.GShl;
    Code[54]:=16-Dst.Bpr;
    Code[58]:=Dst.RShl;
    Code[116]:=Dst.GShl;
    Code[120]:=Dst.GShl;
    Code[127]:=Dst.GShl;
    Code[140]:=Dst.RShl;
    Code[144]:=Dst.RShl;
    Code[151]:=Dst.RShl;
  end else
  begin
    GetMem(Code,SizeOf(TSubMem16REG));
    PSubMem16REG(Code)^:=SubMem16REG;

    i:=Dst.BMask shl 16 or Dst.BMask;
    PDWord(@Code[37])^:=i;
    PDWord(@Code[43])^:=i;
    i:=(i+i)and(not i);
    PDWord(@Code[49])^:=i;
    PDWord(@Code[59])^:=i;
    i:=Dst.GMask shl 16 or Dst.GMask;
    PDWord(@Code[77])^:=i;
    PDWord(@Code[83])^:=i;
    i:=i shr Dst.GShl;
    i:=(i+i)and(not i);
    PDWord(@Code[95])^:=i;
    PDWord(@Code[105])^:=i;
    i:=Dst.RMask shl 16 or Dst.RMask;
    PDWord(@Code[128])^:=i;
    PDWord(@Code[134])^:=i;
    i:=i shr Dst.RShl;
    i:=(i+i)and(not i);
    PDWord(@Code[146])^:=i;
    PDWord(@Code[156])^:=i;
    Code[65]:=Dst.Bpb;
    Code[68]:=(1 shl Dst.Bpb)-1;
    Code[89]:=Dst.GShl;
    Code[92]:=Dst.GShl;
    Code[111]:=Dst.Bpg;
    Code[114]:=(1 shl Dst.Bpg)-1;
    Code[119]:=Dst.GShl;
    Code[140]:=Dst.RShl;
    Code[143]:=Dst.RShl;
    Code[162]:=Dst.Bpr;
    Code[165]:=(1 shl Dst.Bpr)-1;
    Code[170]:=Dst.RShl;
  end;

  for i:=0 to Dst.AbsHeight-1 do
    TSubMem16Proc(Code)(Dst.Scanlines[i],Src1.Scanlines[i],Src2.Scanlines[i],Dst.BWidth-Dst.Gap);

  FreeMem(Code);
end;

procedure SubMem8(Dst,Src1,Src2:Pointer;Size:Integer);
// Dst  = eax
// Src1 = edx
// Src2 = ecx
// Size = [ebp + 8]
// s    = [ebp - 4]
var
  s: Integer;
asm
  push ebx
  push edi
  push esi

  mov edi,edx
  mov esi,ecx
  mov ecx,[ebp+8]
  push ecx
  mov s,esp
  mov esp,eax
  and ecx,$FFFFFFFC
  jz  @skip
  add ecx,eax
  mov [ebp+8],ecx

  @dwords:
    mov  ecx,[edi]
    mov  ebx,[esi]
    mov  eax,ecx
    mov  edx,ebx

    and  ecx,$00FF00FF
    and  ebx,$00FF00FF
    or   ecx,$01000100
    sub  ecx,ebx
    mov  ebx,ecx
    and  ebx,$01000100
    imul ebx,$FF
    shr  ebx,8
    and  ecx,ebx

    and  eax,$FF00FF00
    and  edx,$FF00FF00
    shr  eax,8
    shr  edx,8
    or   eax,$01000100
    sub  eax,edx
    mov  edx,eax
    and  edx,$01000100
    imul edx,$FF
    shr  edx,8
    and  eax,edx
    shl  eax,8
    or   ecx,eax

    mov [esp],ecx
    add edi,4
    add esi,4
    add esp,4
    cmp esp,[ebp+8]
  jne @dwords

  @skip:
  mov eax,esp
  mov esp,s
  pop ecx
  and ecx,11b
  jz  @exit

  @bytes:
    movzx ebx,Byte([edi])
    movzx edx,Byte([esi])
    sub ebx,edx
    mov edx,ebx
    shr edx,8
    xor edx,-1
    and ebx,edx
    mov [eax],bl
    inc edi
    inc esi
    inc eax
    dec ecx
  jnz @bytes

  @exit:
  pop esi
  pop edi
  pop ebx
end;

procedure SubMem8MMX(Dst,Src1,Src2:Pointer;Size:Integer);
// Dst  = eax
// Src1 = edx
// Src2 = ecx
// Size = [ebp + 8]
asm
  push ebx
  push edi
  push esi

  mov edi,edx
  mov esi,ecx
  mov ecx,[ebp+8]
  push ecx
  shr ecx,3
  jz  @skip

  @quads:
    db $0F,$6F,$07 /// movq    mm0,[edi]
    db $0F,$6F,$0E /// movq    mm1,[esi]
    db $0F,$D8,$C1 /// psubusb mm0,mm1
    db $0F,$7F,$00 /// movq    [eax],mm0
    add edi,8
    add esi,8
    add eax,8
    dec ecx
  jnz @quads

  db $0F,$77 // emms

  @skip:
  pop ecx
  and ecx,111b
  jz  @exit

  @bytes:
    movzx ebx,Byte([edi])
    movzx edx,Byte([esi])
    sub ebx,edx
    mov edx,ebx
    shr edx,8
    xor edx,-1
    and ebx,edx
    mov [eax],bl
    inc edi
    inc esi
    inc eax
    dec ecx
  jnz @bytes

  @exit:
  pop esi
  pop edi
  pop ebx
end;

procedure SubBlend8(Dst,Src1,Src2:TFastDIB);
type
  TSubMem8Proc = procedure(Dst,Src1,Src2:Pointer;Size:Integer);
var
  i: Integer;
  SubMem: TSubMem8Proc;
begin
  if cfMMX in CPUInfo.Features then
    SubMem:=SubMem8MMX else SubMem:=SubMem8;

  for i:=0 to Dst.AbsHeight-1 do
    SubMem(Dst.Scanlines[i],Src1.Scanlines[i],Src2.Scanlines[i],Dst.BWidth-Dst.Gap);
end;

procedure SubBlend(Dst,Src1,Src2:TFastDIB);
begin
  case Dst.Bpp of
    8,24,32: SubBlend8(Dst,Src1,Src2);
    16: SubBlend16(Dst,Src1,Src2);
  end;
end;

{
procedure AddMem16(Dst,Src1,Src2:Pointer;Size:Integer);
// Dst  = eax
// Src1 = edx
// Src2 = ecx
// Size = [ebp +  8]
// a    = [ebp -  4]
// b    = [ebp -  8]
// c    = [ebp - 12]
var
  a,b,c: Integer;
asm
  push ebx
  push edi
  push esi

  mov  edi,edx
  mov  esi,ecx
  mov  ecx,[ebp+8]
  push ecx
  and  ecx,$FFFFFFFC
  jz   @skip

  add ecx,eax
  mov [ebp+8],ecx

  @dwords:
    mov ecx,[edi]
    mov ebx,[esi]
    and ecx,$001F001F
    and ebx,$001F001F
    add ecx,ebx
    mov ebx,ecx
    and ebx,$00200020
    shr ebx,5
    imul ebx,$1F
    or  ecx,ebx

    mov edx,[edi]
    mov ebx,[esi]
    and edx,$07E007E0
    and ebx,$07E007E0
    shr edx,5
    shr ebx,5
    add edx,ebx
    mov ebx,edx
    and ebx,$00400040
    shr ebx,6
    imul ebx,$3F
    or  edx,ebx
    shl edx,5
    or  ecx,edx

    mov edx,[edi]
    mov ebx,[esi]
    and edx,$F800F800
    and ebx,$F800F800
    shr edx,11
    shr ebx,11
    add edx,ebx
    mov ebx,edx
    and ebx,$00200020
    shr ebx,5
    imul ebx,$1F
    or  edx,ebx
    shl edx,11
    or  ecx,edx

    mov [eax],ecx
    add edi,4
    add esi,4
    add eax,4
    cmp eax,[ebp+8]
  jne @dwords

  cmp edi,ebp
  je  @last

  @skip:
  pop ecx
  and ecx,11b
  shr ecx,1
  jz  @exit

  mov  cx,[edi]
  mov  bx,[esi]
  mov  a,ecx
  mov  b,ebx
  lea  edi,a
  lea  esi,b
  push eax
  lea  eax,c
  lea  ebx,[eax+4]
  mov  [ebp+8],ebx
  jmp  @dwords

  @last:
  pop eax
  mov ebx,[ebp-12]
  mov [eax],bx

  @exit:
  pop esi
  pop edi
  pop ebx
end;

procedure AddMem16MMX(Dst,Src1,Src2:Pointer;Size:Integer);
// Dst   = eax
// Src1  = edx
// Src2  = ecx
// Size  = [ebp +  8]
// B1:B2 = [ebp -  8]
// G1:G2 = [ebp - 16]
// R1:R2 = [ebp - 24]
// M1:M2 = [ebp - 32]
var
  B1,B2,G1,G2,R1,R2,M1,M2: Integer;
asm
  push edi
  push esi

  mov B1,$001F001F
  mov B2,$001F001F
  mov G1,$07E007E0
  mov G2,$07E007E0
  mov R1,$F800F800
  mov R2,$F800F800
  mov M2,0
  mov M1,0

  movq  mm6,[ebp-16]
  psrlw mm6,5
  movq  mm7,[ebp-24]
  psrlw mm7,11

  mov edi,edx
  mov esi,ecx
  mov ecx,[ebp+8]
  push ecx
  shr ecx,3
  jz  @skip

  @quads:
    movq    mm0,[edi]
    movq    mm1,[esi]
  @cleanup:
    movq    mm2,mm0
    movq    mm3,mm1
    movq    mm4,mm0
    movq    mm5,mm1
    pand    mm0,[ebp-8]
    pand    mm1,[ebp-8]
    paddusw mm0,mm1
    movq    mm1,mm0
    pcmpgtw mm1,[ebp-8]
    psrlw   mm1,11
    por     mm0,mm1

    pand    mm2,[ebp-16]
    pand    mm3,[ebp-16]
    psrlw   mm2,5
    psrlw   mm3,5
    paddusw mm2,mm3
    movq    mm3,mm2
    pcmpgtw mm3,mm6
    psrlw   mm3,10
    por     mm2,mm3
    psllw   mm2,5
    por     mm0,mm2

    pand    mm4,[ebp-24]
    pand    mm5,[ebp-24]
    psrlw   mm4,11
    psrlw   mm5,11
    paddusw mm4,mm5
    movq    mm5,mm4
    pcmpgtw mm5,mm7
    psrlw   mm5,11
    por     mm4,mm5
    psllw   mm4,11
    por     mm0,mm4
    movq    [eax],mm0

    add edi,8
    add esi,8
    add eax,8
    dec ecx
  jnz @quads

  cmp edi,0
  je @last

  @skip:
  pop ecx
  and ecx,111b
  shr ecx,1
  jz  @exit

  mov  edx,ecx
  push edi
  lea  edi,M1
  rep  movsw
  movq mm1,[ebp-32]
  pop  esi
  mov  ecx,edx
  lea  edi,M1
  rep  movsw
  movq mm0,[ebp-32]
  push eax
  lea  eax,M1
  mov  edi,-8
  mov  ecx,1
  jmp  @cleanup

  @last:
  pop edi
  lea esi,M1
  mov ecx,edx
  rep movsw

  @exit:
  emms
  pop esi
  pop edi
end;
}

procedure AddBlend16(Dst,Src1,Src2:TFastDIB);
type
  TAddMem16REG = array[0..235]of Byte; PAddMem16REG =^TAddMem16REG;
  TAddMem16MMX = array[0..346]of Byte; PAddMem16MMX =^TAddMem16MMX;
  TAddMemProc = procedure(Dst,Src1,Src2:Pointer;Size:Integer);
const
  AddMem16REG: TAddMem16REG =
  ($55,$8B,$EC,$83,$C4,$F4,$53,$57,$56,$89,$D7,$89,$CE,$8B,$4D,$08,$51,$83,$E1,
   $FC,$0F,$84,$99,$00,$00,$00,$01,$C1,$89,$4D,$08,$8B,$0F,$8B,$1E,$81,$E1,$1F,
   $00,$1F,$00,$81,$E3,$1F,$00,$1F,$00,$01,$D9,$89,$CB,$81,$E3,$20,$00,$20,$00,
   $C1,$EB,$05,$6B,$DB,$1F,$09,$D9,$8B,$17,$8B,$1E,$81,$E2,$E0,$07,$E0,$07,$81,
   $E3,$E0,$07,$E0,$07,$C1,$EA,$05,$C1,$EB,$05,$01,$DA,$89,$D3,$81,$E3,$40,$00,
   $40,$00,$C1,$EB,$06,$6B,$DB,$3F,$09,$DA,$C1,$E2,$05,$09,$D1,$8B,$17,$8B,$1E,
   $81,$E2,$00,$F8,$00,$F8,$81,$E3,$00,$F8,$00,$F8,$C1,$EA,$0B,$C1,$EB,$0B,$01,
   $DA,$89,$D3,$81,$E3,$20,$00,$20,$00,$C1,$EB,$05,$6B,$DB,$1F,$09,$DA,$C1,$E2,
   $0B,$09,$D1,$89,$08,$83,$C7,$04,$83,$C6,$04,$83,$C0,$04,$3B,$45,$08,$0F,$85,
   $70,$FF,$FF,$FF,$39,$EF,$74,$29,$59,$83,$E1,$03,$D1,$E9,$74,$28,$66,$8B,$0F,
   $66,$8B,$1E,$89,$4D,$FC,$89,$5D,$F8,$8D,$7D,$FC,$8D,$75,$F8,$50,$8D,$45,$F4,
   $8D,$58,$04,$89,$5D,$08,$E9,$43,$FF,$FF,$FF,$58,$8B,$5D,$F4,$66,$89,$18,$5E,
   $5F,$5B,$8B,$E5,$5D,$C2,$04,$00);

  AddMem16MMX: TAddMem16MMX =
  ($55,$8B,$EC,$83,$C4,$E0,$57,$56,$C7,$45,$FC,$1F,$00,$1F,$00,$C7,$45,$F8,$1F,
   $00,$1F,$00,$C7,$45,$F4,$E0,$07,$E0,$07,$C7,$45,$F0,$E0,$07,$E0,$07,$C7,$45,
   $EC,$00,$F8,$00,$F8,$C7,$45,$E8,$00,$F8,$00,$F8,$C7,$45,$E4,$00,$00,$00,$00,
   $C7,$45,$E0,$00,$00,$00,$00,$0F,$6F,$B5,$F0,$FF,$FF,$FF,$0F,$71,$D6,$05,$0F,
   $6F,$BD,$E8,$FF,$FF,$FF,$0F,$71,$D7,$0B,$89,$D7,$89,$CE,$8B,$4D,$08,$51,$C1,
   $E9,$03,$0F,$84,$A6,$00,$00,$00,$0F,$6F,$07,$0F,$6F,$0E,$0F,$6F,$D0,$0F,$6F,
   $D9,$0F,$6F,$E0,$0F,$6F,$E9,$0F,$DB,$85,$F8,$FF,$FF,$FF,$0F,$DB,$8D,$F8,$FF,
   $FF,$FF,$0F,$DD,$C1,$0F,$6F,$C8,$0F,$65,$8D,$F8,$FF,$FF,$FF,$0F,$71,$D1,$0B,
   $0F,$EB,$C1,$0F,$DB,$95,$F0,$FF,$FF,$FF,$0F,$DB,$9D,$F0,$FF,$FF,$FF,$0F,$71,
   $D2,$05,$0F,$71,$D3,$05,$0F,$DD,$D3,$0F,$6F,$DA,$0F,$65,$DE,$0F,$71,$D3,$0A,
   $0F,$EB,$D3,$0F,$71,$F2,$05,$0F,$EB,$C2,$0F,$DB,$A5,$E8,$FF,$FF,$FF,$0F,$DB,
   $AD,$E8,$FF,$FF,$FF,$0F,$71,$D4,$0B,$0F,$71,$D5,$0B,$0F,$DD,$E5,$0F,$6F,$EC,
   $0F,$65,$EF,$0F,$71,$D5,$0B,$0F,$EB,$E5,$0F,$71,$F4,$0B,$0F,$EB,$C4,$0F,$7F,
   $00,$83,$C7,$08,$83,$C6,$08,$83,$C0,$08,$49,$0F,$85,$5F,$FF,$FF,$FF,$83,$FF,
   $00,$74,$3B,$59,$83,$E1,$07,$D1,$E9,$74,$3C,$89,$CA,$57,$8D,$7D,$E0,$66,$F3,
   $A5,$0F,$6F,$8D,$E0,$FF,$FF,$FF,$5E,$89,$D1,$8D,$7D,$E0,$66,$F3,$A5,$0F,$6F,
   $85,$E0,$FF,$FF,$FF,$50,$8D,$45,$E0,$BF,$F8,$FF,$FF,$FF,$B9,$01,$00,$00,$00,
   $E9,$25,$FF,$FF,$FF,$5F,$8D,$75,$E0,$89,$D1,$66,$F3,$A5,$0F,$77,$5E,$5F,$8B,
   $E5,$5D,$C2,$04,$00);
var
  i: Integer;
  Code: PLine8;
begin
  if cfMMX in CPUInfo.Features then
  begin
    GetMem(Code,SizeOf(TAddMem16MMX));
    PAddMem16MMX(Code)^:=AddMem16MMX;

    i:=Dst.BMask shl 16 or Dst.BMask;
    PDWord(@Code[11])^:=i;
    PDWord(@Code[18])^:=i;
    i:=Dst.GMask shl 16 or Dst.GMask;
    PDWord(@Code[25])^:=i;
    PDWord(@Code[32])^:=i;
    i:=Dst.RMask shl 16 or Dst.RMask;
    PDWord(@Code[39])^:=i;
    PDWord(@Code[46])^:=i;
    Code[74]:=Dst.GShl;
    Code[85]:=Dst.RShl;
    Code[151]:=16-Dst.Bpb;
    Code[172]:=Dst.GShl;
    Code[176]:=Dst.GShl;
    Code[189]:=16-Dst.Bpg;
    Code[196]:=Dst.GShl;
    Code[217]:=Dst.RShl;
    Code[221]:=Dst.RShl;
    Code[234]:=16-Dst.Bpr;
    Code[241]:=Dst.RShl;
  end else
  begin
    GetMem(Code,SizeOf(TAddMem16REG));
    PAddMem16REG(Code)^:=AddMem16REG;

    i:=Dst.BMask shl 16 or Dst.BMask;
    PDWord(@Code[37])^:=i;
    PDWord(@Code[43])^:=i;
    PDWord(@Code[53])^:=(i+i)and(not i);
    i:=Dst.GMask shl 16 or Dst.GMask;
    PDWord(@Code[71])^:=i;
    PDWord(@Code[77])^:=i;
    i:=i shr Dst.GShl;
    PDWord(@Code[93])^:=(i+i)and(not i);
    i:=Dst.RMask shl 16 or Dst.RMask;
    PDWord(@Code[116])^:=i;
    PDWord(@Code[122])^:=i;
    i:=i shr Dst.RShl;
    PDWord(@Code[138])^:=(i+i)and(not i);
    Code[59]:=Dst.Bpb;
    Code[62]:=(1 shl Dst.Bpb)-1;
    Code[83]:=Dst.GShl;
    Code[86]:=Dst.GShl;
    Code[99]:=Dst.Bpg;
    Code[102]:=(1 shl Dst.Bpg)-1;
    Code[107]:=Dst.GShl;
    Code[128]:=Dst.RShl;
    Code[131]:=Dst.RShl;
    Code[144]:=Dst.Bpr;
    Code[147]:=(1 shl Dst.Bpr)-1;
    Code[152]:=Dst.RShl;
  end;

  for i:=0 to Dst.AbsHeight-1 do
    TAddMemProc(Code)(Dst.Scanlines[i],Src1.Scanlines[i],Src2.Scanlines[i],Dst.BWidth-Dst.Gap);

  FreeMem(Code);
end;

procedure AddMem8(Dst,Src1,Src2:Pointer;Size:Integer);
// Dst  = eax
// Src1 = edx
// Src2 = ecx
// Size = [ebp + 8]
var
  s: Integer;
asm
  push ebx
  push edi
  push esi

  mov edi,edx
  mov esi,ecx
  mov ecx,[ebp+8]
  push ecx
  mov s,esp
  mov esp,eax
  and ecx,$FFFFFFFC
  jz @skip

  add ecx,eax
  mov [ebp+8],ecx

  @dwords:
    mov eax,[edi]
    mov ebx,[esi]
    mov ecx,eax
    mov edx,ebx

    and eax,$00FF00FF
    and ebx,$00FF00FF
    add eax,ebx
    mov ebx,eax
    and ebx,$01000100
    shr ebx,8
    imul ebx,$FF
    or  eax,ebx

    and ecx,$FF00FF00
    and edx,$FF00FF00
    shr ecx,8
    shr edx,8
    add ecx,edx
    mov edx,ecx
    and edx,$01000100
    shr edx,8
    imul edx,$FF
    or  ecx,edx

    shl ecx,8
    or eax,ecx
    mov [esp],eax

    add edi,4
    add esi,4
    add esp,4
    cmp esp,[ebp+8]
  jne @dwords

  @skip:
  mov eax,esp
  mov esp,s
  pop ecx
  and ecx,11b
  jz @exit

  @bytes:
    movzx ebx,Byte([esi])
    movzx edx,Byte([edi])
    add ebx,edx
    mov edx,ebx
    and edx,$0100
    sub edx,$0100
    xor edx,-1
    shr edx,8
    or  ebx,edx
    mov [eax],bl
    inc esi
    inc edi
    inc eax
    dec ecx
  jnz @bytes

  @exit:
  pop esi
  pop edi
  pop ebx
end;

procedure AddMem8MMX(Dst,Src1,Src2:Pointer;Size:Integer);
// Dst  = eax
// Src1 = edx
// Src2 = ecx
// Size = [ebp + 8]
asm
  push ebx
  push edi
  push esi

  mov edi,edx
  mov esi,ecx
  mov ecx,[ebp+8]
  push ecx
  shr ecx,3
  jz @skip

  @quads:
    db $0F,$6F,$07  /// movq    mm0,[edi]
    db $0F,$6F,$0E  /// movq    mm1,[esi]
    db $0F,$DC,$C1  /// paddusb mm0,mm1
    db $0F,$7F,$00  /// movq    [eax],mm0
    add edi,8
    add esi,8
    add eax,8
    dec ecx
  jnz @quads

  db $0F,$77 // emms

  @skip:
  pop ecx
  and ecx,111b
  jz  @exit

  @bytes:
    movzx ebx,Byte([esi])
    movzx edx,Byte([edi])
    add ebx,edx
    mov edx,ebx
    and edx,$0100
    sub edx,$0100
    xor edx,-1
    shr edx,8
    or  ebx,edx
    mov [eax],bl

    inc edi
    inc esi
    inc eax
    dec ecx
  jnz @bytes

  @exit:
  pop esi
  pop edi
  pop ebx
end;

procedure AddBlend8(Dst,Src1,Src2:TFastDIB);
type
  TAddMemProc = procedure(Dst,Src1,Src2:Pointer;Size:Integer);
var
  i: Integer;
  AddMem: TAddMemProc;
begin
  if cfMMX in CPUInfo.Features then
    AddMem:=AddMem8MMX else AddMem:=AddMem8;

  for i:=0 to Dst.AbsHeight-1 do
    AddMem(Dst.Scanlines[i],Src1.Scanlines[i],Src2.Scanlines[i],Dst.BWidth-Dst.Gap);
end;

procedure AddBlend(Dst,Src1,Src2:TFastDIB);
begin
  case Dst.Bpp of
    8,24,32: AddBlend8(Dst,Src1,Src2);
    16: AddBlend16(Dst,Src1,Src2);
  end;
end;

{
procedure MulMem16(Dst,Src1,Src2:Pointer;Size:Integer);
// Dst  = eax
// Src1 = edx
// Src2 = ecx
// Size = [ebp + 8]
// a    = [ebp - 4]
// b    = [ebp - 8]
var
  a,b,s: Integer;
asm
  push ebx
  push edi
  push esi
  mov s,esp

  mov edi,edx
  mov esi,ecx
  mov esp,eax

  mov ecx,[ebp+8]
  shr ecx,1
  jz @exit

  @words:
    movzx ebx,Word([edi])
    movzx edx,Word([esi])
    mov a,ebx
    mov b,edx

    and ebx,$0000001F
    and edx,$0000001F
    imul ebx,edx
    shr ebx,5

    mov eax,a
    mov edx,b
    and eax,$000007E0
    and edx,$000007E0
    shr eax,5
    shr edx,5
    imul eax,edx
    shr eax,6
    shl eax,5
    or ebx,eax

    mov eax,a
    mov edx,b
    and eax,$0000F800
    and edx,$0000F800
    shr eax,11
    shr edx,11
    imul eax,edx
    shr eax,5
    shl eax,11
    or ebx,eax

    mov [esp],bx
    add edi,2
    add esi,2
    add esp,2
    dec ecx
  jnz @words

  @exit:
  mov esp,s
  pop esi
  pop edi
  pop ebx
end;

procedure MulMem16MMX(Dst,Src1,Src2:Pointer;Size:Integer);
// Dst   = eax
// Src1  = edx
// Src2  = ecx
// Size  = [ebp +  8]
// HB:LB = [ebp -  8]
// HG:LG = [ebp - 16]
// HR:LR = [ebp - 24]
// AA:BB = [ebp - 32]
var
  HB,LB,HG,LG,HR,LR,AA,BB: Integer;
asm
  push edi
  push esi

  mov HR,$F800F800
  mov LR,$F800F800
  mov HG,$07E007E0
  mov LG,$07E007E0
  mov HB,$001F001F
  mov LB,$001F001F
  mov AA,0
  mov BB,0

  mov  edi,edx
  mov  esi,ecx
  mov  ecx,[ebp+8]
  push ecx
  shr  ecx,3
  jz   @skip

  @quads:
    movq   mm0,[edi]
    movq   mm1,[esi]
  @cleanup:
    movq   mm2,mm0
    movq   mm3,mm0
    movq   mm4,mm1
    movq   mm5,mm1
    pand   mm0,[ebp-24]
    pand   mm1,[ebp-24]
    pmullw mm0,mm1
    psrlw  mm0,5
    pand   mm2,[ebp-16]
    pand   mm4,[ebp-16]
    psrlw  mm2,5
    psrlw  mm4,5
    pmullw mm2,mm4
    psrlw  mm2,6
    psllw  mm2,5
    por    mm0,mm2
    pand   mm3,[ebp-8]
    pand   mm5,[ebp-8]
    psrlw  mm3,11
    psrlw  mm5,11
    pmullw mm3,mm5
    psrlw  mm3,5
    psllw  mm3,11
    por    mm0,mm3
    movq   [eax],mm0

    add edi,8
    add esi,8
    add eax,8
    dec ecx
  jnz @quads

  cmp edi,0
  je @last

  @skip:
  pop ecx
  and ecx,111b
  shr ecx,1
  jz  @exit

  mov  edx,ecx
  push edi
  mov  edi,ebp
  sub  edi,32 // edi = ebp-32
  rep  movsw
  movq mm1,[ebp-32]
  pop  esi
  sub  edi,edx // edi = ebp-32
  mov  ecx,edx
  rep  movsw
  movq mm0,[ebp-32]
  mov  ecx,1
  mov  edi,-8
  push eax
  mov  eax,ebp
  sub  eax,32
  jmp  @cleanup

  @last:
  mov ecx,edx
  pop edi
  mov esi,ebp
  sub esi,32
  rep movsw

  @exit:
  emms
  pop esi
  pop edi
end;
}

procedure MulBlend16(Dst,Src1,Src2:TFastDIB);
type
  TMulMem16REG = array[0..150]of Byte; PMulMem16REG =^TMulMem16REG;
  TMulMem16MMX = array[0..294]of Byte; PMulMem16MMX =^TMulMem16MMX;
  TMulMemProc = procedure(Dst,Src1,Src2:Pointer;Size:Integer);
const
  MulMem16REG: TMulMem16REG =
  ($55,$8B,$EC,$83,$C4,$F4,$53,$57,$56,$89,$65,$FC,$89,$D7,$89,$CE,$89,$C4,$8B,
   $4D,$08,$D1,$E9,$74,$72,$0F,$B7,$1F,$0F,$B7,$16,$89,$5D,$F8,$89,$55,$F4,$81,
   $E3,$1F,$00,$00,$00,$81,$E2,$1F,$00,$00,$00,$0F,$AF,$DA,$C1,$EB,$05,$8B,$45,
   $F8,$8B,$55,$F4,$25,$E0,$07,$00,$00,$81,$E2,$E0,$07,$00,$00,$C1,$E8,$05,$C1,
   $EA,$05,$0F,$AF,$C2,$C1,$E8,$06,$C1,$E0,$05,$09,$C3,$8B,$45,$F8,$8B,$55,$F4,
   $25,$00,$F8,$00,$00,$81,$E2,$00,$F8,$00,$00,$C1,$E8,$0B,$C1,$EA,$0B,$0F,$AF,
   $C2,$C1,$E8,$05,$C1,$E0,$0B,$09,$C3,$66,$89,$1C,$24,$83,$C7,$02,$83,$C6,$02,
   $83,$C4,$02,$49,$75,$8E,$8B,$65,$FC,$5E,$5F,$5B,$8B,$E5,$5D,$C2,$04,$00);

  MulMem16MMX: TMulMem16MMX =
  ($55,$8B,$EC,$83,$C4,$E0,$57,$56,$C7,$45,$FC,$00,$F8,$00,$F8,$C7,$45,$F8,$00,
   $F8,$00,$F8,$C7,$45,$F4,$E0,$07,$E0,$07,$C7,$45,$F0,$E0,$07,$E0,$07,$C7,$45,
   $EC,$1F,$00,$1F,$00,$C7,$45,$E8,$1F,$00,$1F,$00,$C7,$45,$E4,$00,$00,$00,$00,
   $C7,$45,$E0,$00,$00,$00,$00,$89,$D7,$89,$CE,$8B,$4D,$08,$51,$C1,$E9,$03,$0F,
   $84,$83,$00,$00,$00,$0F,$6F,$07,$0F,$6F,$0E,$0F,$6F,$D0,$0F,$6F,$D8,$0F,$6F,
   $E1,$0F,$6F,$E9,$0F,$DB,$85,$E8,$FF,$FF,$FF,$0F,$DB,$8D,$E8,$FF,$FF,$FF,$0F,
   $D5,$C1,$0F,$71,$D0,$05,$0F,$DB,$95,$F0,$FF,$FF,$FF,$0F,$DB,$A5,$F0,$FF,$FF,
   $FF,$0F,$71,$D2,$05,$0F,$71,$D4,$05,$0F,$D5,$D4,$0F,$71,$D2,$06,$0F,$71,$F2,
   $05,$0F,$EB,$C2,$0F,$DB,$9D,$F8,$FF,$FF,$FF,$0F,$DB,$AD,$F8,$FF,$FF,$FF,$0F,
   $71,$D3,$0B,$0F,$71,$D5,$0B,$0F,$D5,$DD,$0F,$71,$D3,$05,$0F,$71,$F3,$0B,$0F,
   $EB,$C3,$0F,$7F,$00,$83,$C7,$08,$83,$C6,$08,$83,$C0,$08,$49,$75,$82,$83,$FF,
   $00,$74,$3E,$59,$83,$E1,$07,$D1,$E9,$74,$41,$89,$CA,$57,$89,$EF,$83,$EF,$20,
   $66,$F3,$A5,$0F,$6F,$8D,$E0,$FF,$FF,$FF,$5E,$29,$D7,$89,$D1,$66,$F3,$A5,$0F,
   $6F,$85,$E0,$FF,$FF,$FF,$B9,$01,$00,$00,$00,$BF,$F8,$FF,$FF,$FF,$50,$89,$E8,
   $83,$E8,$20,$E9,$45,$FF,$FF,$FF,$89,$D1,$5F,$89,$EE,$83,$EE,$20,$66,$F3,$A5,
   $0F,$77,$5E,$5F,$8B,$E5,$5D,$C2,$04,$00);
var
  i: Integer;
  Code: PLine8;
begin
  if cfMMX in CPUInfo.Features then
  begin
    GetMem(Code,SizeOf(TMulMem16MMX));
    PMulMem16MMX(Code)^:=MulMem16MMX;

    i:=Dst.RMask shl 16 or Dst.RMask;
    PDWord(@Code[11])^:=i;
    PDWord(@Code[18])^:=i;
    i:=Dst.GMask shl 16 or Dst.GMask;
    PDWord(@Code[25])^:=i;
    PDWord(@Code[32])^:=i;
    i:=Dst.BMask shl 16 or Dst.BMask;
    PDWord(@Code[39])^:=i;
    PDWord(@Code[46])^:=i;
    Code[119]:=Dst.Bpb;
    Code[137]:=Dst.GShl;
    Code[141]:=Dst.GShl;
    Code[148]:=Dst.Bpg;
    Code[152]:=Dst.GShl;
    Code[173]:=Dst.RShl;
    Code[177]:=Dst.RShl;
    Code[184]:=Dst.Bpr;
    Code[188]:=Dst.RShl;
  end else
  begin
    GetMem(Code,SizeOf(TMulMem16REG));
    PMulMem16REG(Code)^:=MulMem16REG;

    PDWord(@Code[39])^:=Dst.BMask;
    PDWord(@Code[45])^:=Dst.BMask;
    PDWord(@Code[62])^:=Dst.GMask;
    PDWord(@Code[68])^:=Dst.GMask;
    PDWord(@Code[96])^:=Dst.RMask;
    PDWord(@Code[102])^:=Dst.RMask;
    Code[54]:=Dst.Bpb;
    Code[74]:=Dst.GShl;
    Code[77]:=Dst.GShl;
    Code[83]:=Dst.Bpg;
    Code[86]:=Dst.GShl;
    Code[108]:=Dst.RShl;
    Code[111]:=Dst.RShl;
    Code[117]:=Dst.Bpr;
    Code[120]:=Dst.RShl;
  end;

  for i:=0 to Dst.AbsHeight-1 do
    TMulMemProc(Code)(Dst.Scanlines[i],Src1.Scanlines[i],Src2.Scanlines[i],Dst.BWidth-Dst.Gap);

  FreeMem(Code);
end;

procedure MulMem8(Dst,Src1,Src2:Pointer;Size:Integer);
// Dst  = eax
// Src1 = edx
// Src2 = ecx
// Size = [ebp + 8]
asm
  push ebx
  push edi
  push esi

  mov edi,edx
  mov esi,ecx
  mov ecx,[ebp+8]
  cmp ecx,0
  je @exit

  @bytes:
    movzx ebx,Byte([edi])
    movzx edx,Byte([esi])
    imul ebx,edx
    shr ebx,8
    mov [eax],bl
    inc edi
    inc esi
    inc eax
    dec ecx
  jnz @bytes

  @exit:
  pop esi
  pop edi
  pop ebx
end;

procedure MulMem8MMX(Dst,Src1,Src2:Pointer;Size:Integer);
// Dst  = eax
// Src1 = edx
// Src2 = ecx
// Size = [ebp + 8]
asm
  push ebx
  push edi
  push esi

  mov edi,edx
  mov esi,ecx
  mov ecx,[ebp+8]
  push ecx
  shr ecx,2
  jz @skip

  db $0F,$EF,$D2       /// pxor mm2,mm2

  @dwords:
    db $0F,$6E,$07       /// movd       mm0,[edi]
    db $0F,$6E,$0E       /// movd       mm1,[esi]
    db $0F,$60,$C2       /// punpcklbw  mm0,mm2
    db $0F,$60,$CA       /// punpcklbw  mm1,mm2
    db $0F,$D5,$C1       /// pmullw     mm0,mm1
    db $0F,$71,$D0,$08   /// psrlw      mm0,8
    db $0F,$67,$C0       /// packuswb   mm0,mm0
    db $0F,$7E,$00       /// movd       [eax],mm0

    add edi,4
    add esi,4
    add eax,4
    dec ecx
  jnz @dwords

  db $0F,$77 // emms

  @skip:
  pop ecx
  and ecx,11b
  jz @exit

  @bytes:
    movzx ebx,Byte([edi])
    movzx edx,Byte([esi])
    imul ebx,edx
    shr ebx,8
    mov [eax],bl
    inc edi
    inc esi
    inc eax
    dec ecx
  jnz @bytes

  @exit:
  pop esi
  pop edi
  pop ebx
end;

procedure MulBlend8(Dst,Src1,Src2:TFastDIB);
type
  TMulMemProc = procedure(Dst,Src1,Src2:Pointer;Size:Integer);
var
  i: Integer;
  MulMem: TMulMemProc;
begin
  if cfMMX in CPUInfo.Features then
    MulMem:=MulMem8MMX else MulMem:=MulMem8;

  for i:=0 to Dst.AbsHeight-1 do
    MulMem(Dst.Scanlines[i],Src1.Scanlines[i],Src2.Scanlines[i],Dst.BWidth-Dst.Gap);
end;

procedure MulBlend(Dst,Src1,Src2:TFastDIB);
begin
  case Dst.Bpp of
    8,24,32: MulBlend8(Dst,Src1,Src2);
    16: MulBlend16(Dst,Src1,Src2);
  end;
end;

{
procedure BlendMem16(Dst,Src1,Src2:Pointer;Size,Alpha:Integer);
// Dst   = eax
// Src1  = edx
// Src2  = ecx
// Size  = [ebp + 12]
// Alpha = [ebp +  8]
// a     = [ebp -  4]
// b     = [ebp -  8]
// c     = [ebp - 12]
var
  a,b,c: Cardinal;
asm
  push ebx
  push edi
  push esi
  mov  edi,edx
  mov  esi,ecx

  mov  ecx,[ebp+12]
  and  ecx,3
  push ecx
  sub  [ebp+12],ecx
  jz   @skip
  add  [ebp+12],eax

  @dwords:
    mov  edx,[edi]     // edx = src1
    mov  ecx,[esi]     // ecx = src2
  @words:              //
    mov  a,edx         // a = copy of src1
    mov  b,ecx         // b = copy of src2
                       //
    and  edx,$001F001F // edx = blue_src1
    and  ecx,$001F001F // ecx = blue_src2
    sub  edx,ecx       // (blue_src1-blue_src2)
    imul edx,[ebp+8]   // (blue_src1-blue_src2)*blue_alpha
    shr  edx,8         // (blue_src1-blue_src2)*blue_alpha / 256
    add  edx,ecx       // (blue_src1-blue_src2)*blue_alpha / 256 + blue_src2
    and  edx,$001F001F // edx = [-----------bbbbb-----------bbbbb]
                       //
    mov  ebx,a         // ebx = src1
    mov  ecx,b         // ecx = src2
    and  ebx,$03E003E0 // ebx = green_src1
    and  ecx,$03E003E0 // ecx = green_src2
    shr  ebx,5         // [------ggggg-----] >> [-----------ggggg]
    shr  ecx,5         // [------ggggg-----] >> [-----------ggggg]
    sub  ebx,ecx       // (green_src1-green_src2)
    imul ebx,[ebp+8]   // (green_src1-green_src2)*green_alpha
    shr  ebx,8         // (green_src1-green_src2)*green_alpha / 256
    add  ebx,ecx       // (green_src1-green_src2)*green_alpha / 256 + green_src2
    shl  ebx,5         // [------ggggg-----] << [-----------ggggg]
    and  ebx,$03E003E0 // ebx = [------ggggg-----------ggggg-----]
    or   edx,ebx       // edx = [------gggggbbbbb------gggggbbbbb]
                       //
    mov  ebx,a         // ebx = src1
    mov  ecx,b         // ecx = src2
    and  ebx,$7C007C00 // ebx = red_src1
    and  ecx,$7C007C00 // ecx = red_src2
    shr  ebx,10        // [-rrrrr----------] >> [-----------rrrrr]
    shr  ecx,10        // [-rrrrr----------] >> [-----------rrrrr]
    sub  ebx,ecx       // (red_src1-red_src2)
    imul ebx,[ebp+8]   // (red_src1-red_src2)*red_alpha
    shr  ebx,8         // (red_src1-red_src2)*red_alpha / 256
    add  ebx,ecx       // (red_src1-red_src2)*red_alpha / 256 + red_src2
    shl  ebx,10        // [-rrrrr----------] << [-----------rrrrr]
    and  ebx,$7C007C00 // ebx = [-rrrrr-----------rrrrr----------]
    or   edx,ebx       // edx = [-rrrrrgggggbbbbb-rrrrrgggggbbbbb]

    mov  [eax],edx

    add  eax,4
    add  edi,4
    add  esi,4
    cmp  eax,[ebp+12]
  jne @dwords

  mov  ebx,ebp
  sub  ebx,8
  cmp  eax,ebx
  je   @last //eax = @c (ebp-12)

  @skip:
  pop  ecx
  and  ecx,2
  jz   @exit

  push eax
  mov  ebx,ebp
  sub  ebx,12
  mov  eax,ebx //eax = @c (ebp-12)
  add  ebx,4
  mov  [ebp+12],ebx
  mov  dx,[edi]
  mov  cx,[esi]
  jmp  @words

  @last:
  mov  ebx,c
  pop  eax
  mov  [eax],bx

  @exit:
  pop esi
  pop edi
  pop ebx
end;

procedure BlendMem16MMX(Dst,Src1,Src2:Pointer;Size,Alpha:Integer);
// Dst   = eax
// Src1  = edx
// Src2  = ecx
// Size  = [ebp + 12]
// Alpha = [ebp +  8]
// HA:LA = [ebp -  8]
// HB:LB = [ebp - 16]
var
  HA,LA,HB,LB: Integer;
asm
  push ebx
  push edi
  push esi

  mov edi,edx
  mov esi,ecx
  mov ecx,[ebp+12]
  shr ecx,3
  jz  @skip

  mov    ebx,[ebp+8]
  shl    ebx,16
  or     ebx,[ebp+8]
  mov    HA,ebx
  mov    LA,ebx   // HA:LA = 00aa00aa00aa00aa
  mov    HB,0
  mov    LB,0
  mov    ebx,$001F001F
  movd   mm2,ebx
  movd   mm3,ebx
  psllq  mm2,32
  por    mm2,mm3  // mm2 = 001F001F001F001F
  mov    ebx,$03E003E0
  movd   mm3,ebx
  movd   mm4,ebx
  psllq  mm3,32
  por    mm3,mm4  // mm3 = 03E003E003E003E0
  mov    ebx,$7C007C00
  movd   mm4,ebx
  movd   mm5,ebx
  psllq  mm4,32
  por    mm4,mm5  // mm4 = 7C007C007C007C00

  @quads:
    movq   mm7,[edi]
    movq   mm1,[esi]
  @words:
    movq   mm0,mm7
    movq   mm5,mm1
    pand   mm7,mm2
    pand   mm1,mm2

    psubw  mm7,mm1
    pmullw mm7,[ebp-8]
    psrlw  mm7,8
    paddb  mm7,mm1

    movq   mm1,mm0
    movq   mm6,mm5
    pand   mm0,mm3
    pand   mm5,mm3
    psrlw  mm0,5
    psrlw  mm5,5
    psubw  mm0,mm5
    pmullw mm0,[ebp-8]
    psrlw  mm0,8
    paddb  mm0,mm5
    psllw  mm0,5
    por    mm7,mm0

    pand   mm1,mm4
    pand   mm6,mm4
    psrlw  mm1,10
    psrlw  mm6,10
    psubw  mm1,mm6
    pmullw mm1,[ebp-8]
    psrlw  mm1,8
    paddb  mm1,mm6
    psllw  mm1,10
    por    mm7,mm1
    movq   [eax],mm7

    add eax,8
    add edi,8
    add esi,8
    dec ecx
  jnz @quads

  mov ebx,ebp
  sub ebx,8
  cmp eax,ebx
  je  @last

  @skip:
  mov ecx,[ebp+12]
  and ecx,7
  shr ecx,1
  jz  @exit

  push ecx
  push edi
  push ecx
  mov  edi,ebp
  sub  edi,16
  rep  movsw
  movq mm1,[ebp-16]
  pop  ecx
  pop  esi
  mov  edi,ebp
  sub  edi,16
  rep  movsw
  movq mm7,[ebp-16]
  push eax
  mov  eax,ebp
  sub  eax,16
  mov  ecx,1
  jmp  @words

  @last:
  pop edi
  pop ecx
  mov esi,ebp
  sub esi,16
  rep movsw

  @exit:
  pop esi
  pop edi
  pop ebx

  emms
end;
}

procedure AlphaBlend16(Dst,Src1,Src2:TFastDIB;Alpha:Integer);
type
  TBlendMem16REG = array[0..238]of Byte; PBlendMem16REG =^TBlendMem16REG;
  TBlendMem16MMX = array[0..334]of Byte; PBlendMem16MMX =^TBlendMem16MMX;
  TBlendMemProc = procedure(Dst,Src1,Src2:Pointer;Size,Alpha:Integer);
const
  BlendMem16REG: TBlendMem16REG =
  ($55,$8B,$EC,$83,$C4,$F4,$53,$57,$56,$89,$D7,$89,$CE,$8B,$4D,$0C,$83,$E1,$03,$51,
   $29,$4D,$0C,$0F,$84,$A3,$00,$00,$00,$01,$45,$0C,$8B,$17,$8B,$0E,$89,$55,$FC,$89,
   $4D,$F8,$81,$E2,$1F,$00,$1F,$00,$81,$E1,$1F,$00,$1F,$00,$29,$CA,$0F,$AF,$55,$08,
   $C1,$EA,$08,$01,$CA,$81,$E2,$1F,$00,$1F,$00,$8B,$5D,$FC,$8B,$4D,$F8,$81,$E3,$E0,
   $03,$E0,$03,$81,$E1,$E0,$03,$E0,$03,$C1,$EB,$05,$C1,$E9,$05,$29,$CB,$0F,$AF,$5D,
   $08,$C1,$EB,$08,$01,$CB,$C1,$E3,$05,$81,$E3,$E0,$03,$E0,$03,$09,$DA,$8B,$5D,$FC,
   $8B,$4D,$F8,$81,$E3,$00,$7C,$00,$7C,$81,$E1,$00,$7C,$00,$7C,$C1,$EB,$0A,$C1,$E9,
   $0A,$29,$CB,$0F,$AF,$5D,$08,$C1,$EB,$08,$01,$CB,$C1,$E3,$0A,$81,$E3,$00,$7C,$00,
   $7C,$09,$DA,$89,$10,$83,$C0,$04,$83,$C7,$04,$83,$C6,$04,$3B,$45,$0C,$0F,$85,$69,
   $FF,$FF,$FF,$89,$EB,$83,$EB,$08,$39,$D8,$74,$1F,$59,$83,$E1,$02,$74,$20,$50,$89,
   $EB,$83,$EB,$0C,$89,$D8,$83,$C3,$04,$89,$5D,$0C,$66,$8B,$17,$66,$8B,$0E,$E9,$45,
   $FF,$FF,$FF,$8B,$5D,$F4,$58,$66,$89,$18,$5E,$5F,$5B,$8B,$E5,$5D,$C2,$08,$00);

  BlendMem16MMX: TBlendMem16MMX =
  ($55,$8B,$EC,$83,$C4,$F0,$53,$57,$56,$89,$D7,$89,$CE,$8B,$4D,$0C,$C1,$E9,$03,$0F,
   $84,$E4,$00,$00,$00,$8B,$5D,$08,$C1,$E3,$10,$0B,$5D,$08,$89,$5D,$FC,$89,$5D,$F8,
   $C7,$45,$F4,$00,$00,$00,$00,$C7,$45,$F0,$00,$00,$00,$00,$BB,$1F,$00,$1F,$00,$0F,
   $6E,$D3,$0F,$6E,$DB,$0F,$73,$F2,$20,$0F,$EB,$D3,$BB,$E0,$03,$E0,$03,$0F,$6E,$DB,
   $0F,$6E,$E3,$0F,$73,$F3,$20,$0F,$EB,$DC,$BB,$00,$7C,$00,$7C,$0F,$6E,$E3,$0F,$6E,
   $EB,$0F,$73,$F4,$20,$0F,$EB,$E5,$0F,$6F,$3F,$0F,$6F,$0E,$0F,$6F,$C7,$0F,$6F,$E9,
   $0F,$DB,$FA,$0F,$DB,$CA,$0F,$F9,$F9,$0F,$D5,$BD,$F8,$FF,$FF,$FF,$0F,$71,$D7,$08,
   $0F,$FC,$F9,$0F,$6F,$C8,$0F,$6F,$F5,$0F,$DB,$C3,$0F,$DB,$EB,$0F,$71,$D0,$05,$0F,
   $71,$D5,$05,$0F,$F9,$C5,$0F,$D5,$85,$F8,$FF,$FF,$FF,$0F,$71,$D0,$08,$0F,$FC,$C5,
   $0F,$71,$F0,$05,$0F,$EB,$F8,$0F,$DB,$CC,$0F,$DB,$F4,$0F,$71,$D1,$0A,$0F,$71,$D6,
   $0A,$0F,$F9,$CE,$0F,$D5,$8D,$F8,$FF,$FF,$FF,$0F,$71,$D1,$08,$0F,$FC,$CE,$0F,$71,
   $F1,$0A,$0F,$EB,$F9,$0F,$7F,$38,$83,$C0,$08,$83,$C7,$08,$83,$C6,$08,$49,$0F,$85,
   $78,$FF,$FF,$FF,$89,$EB,$83,$EB,$08,$39,$D8,$74,$3D,$8B,$4D,$0C,$83,$E1,$07,$D1,
   $E9,$74,$3D,$51,$57,$51,$89,$EF,$83,$EF,$10,$66,$F3,$A5,$0F,$6F,$8D,$F0,$FF,$FF,
   $FF,$59,$5E,$89,$EF,$83,$EF,$10,$66,$F3,$A5,$0F,$6F,$BD,$F0,$FF,$FF,$FF,$50,$89,
   $E8,$83,$E8,$10,$B9,$01,$00,$00,$00,$E9,$38,$FF,$FF,$FF,$5F,$59,$89,$EE,$83,$EE,
   $10,$66,$F3,$A5,$5E,$5F,$5B,$0F,$77,$8B,$E5,$5D,$C2,$08,$00);
var
  i: Integer;
  Code: PLine8;
begin
  if cfMMX in CPUInfo.Features then
  begin
    GetMem(Code,SizeOf(BlendMem16MMX));
    PBlendMem16MMX(Code)^:=BlendMem16MMX;

    PDWord(@Code[55])^:=Dst.BMask shl 16 or Dst.BMask;
    PDWord(@Code[73])^:=Dst.GMask shl 16 or Dst.GMask;
    PDWord(@Code[91])^:=Dst.RMask shl 16 or Dst.RMask;
    Code[158]:=Dst.GShl;
    Code[162]:=Dst.GShl;
    Code[183]:=Dst.GShl;
    Code[196]:=Dst.RShl;
    Code[200]:=Dst.RShl;
    Code[221]:=Dst.RShl;
  end else
  begin
    GetMem(Code,SizeOf(BlendMem16REG));
    PBlendMem16REG(Code)^:=BlendMem16REG;

    i:=Dst.BMask shl 16 or Dst.BMask;
    PDWord(@Code[44])^:=i;
    PDWord(@Code[50])^:=i;
    PDWord(@Code[67])^:=i;
    i:=Dst.GMask shl 16 or Dst.GMask;
    PDWord(@Code[79])^:=i;
    PDWord(@Code[85])^:=i;
    PDWord(@Code[111])^:=i;
    i:=Dst.RMask shl 16 or Dst.RMask;
    PDWord(@Code[125])^:=i;
    PDWord(@Code[131])^:=i;
    PDWord(@Code[157])^:=i;
    Code[91]:=Dst.GShl;
    Code[94]:=Dst.GShl;
    Code[108]:=Dst.GShl;
    Code[137]:=Dst.RShl;
    Code[140]:=Dst.RShl;
    Code[154]:=Dst.RShl;
  end;

  for i:=0 to Dst.AbsHeight-1 do
    TBlendMemProc(Code)(Dst.Scanlines[i],Src1.Scanlines[i],Src2.Scanlines[i],Dst.BWidth-Dst.Gap,Alpha);

  FreeMem(Code);
end;

procedure BlendMem8(Dst,Src1,Src2:Pointer;Size,Alpha:Integer);
// Dst   = eax
// Src1  = edx
// Src2  = ecx
// Size  = [ebp + 12]
// Alpha = [ebp +  8]
asm
  push ebx
  push edi
  push esi

  mov  edi,edx
  mov  esi,ecx

  mov  ebx,[ebp+12]
  and  ebx,3
  push ebx
  sub  [ebp+12],ebx
  jz   @skip
  add  [ebp+12],eax

  @dwords:
    mov  ebx,[edi]
    mov  edx,[esi]
    and  ebx,$00FF00FF
    and  edx,$00FF00FF
    sub  ebx,edx
    imul ebx,[ebp+8]
    shr  ebx,8
    add  ebx,edx
    and  ebx,$00FF00FF

    mov  ecx,[edi]
    mov  edx,[esi]
    and  ecx,$FF00FF00
    and  edx,$FF00FF00
    shr  ecx,8
    shr  edx,8
    sub  ecx,edx
    imul ecx,[ebp+8]
    shr  ecx,8
    add  ecx,edx
    shl  ecx,8
    and  ecx,$FF00FF00
    or   ecx,ebx
    mov  [eax],ecx

    add  eax,4
    add  esi,4
    add  edi,4
    cmp  eax,[ebp+12]
  jne @dwords

  @skip:
  pop ebx
  cmp ebx,0
  je  @exit
  add ebx,eax
  mov [ebp+12],ebx

  @bytes:
    mov  bl,[edi]
    mov  cl,[esi]
    and  ebx,$FF
    and  ecx,$FF
    sub  ebx,ecx
    imul ebx,[ebp+8]
    shr  ebx,8
    add  ebx,ecx
    mov  [eax],bl
    inc  eax
    inc  esi
    inc  edi
    cmp  eax,[ebp+12]
  jne @bytes

  @exit:
  pop esi
  pop edi
  pop ebx
end;

procedure BlendMem8MMX(Dst,Src1,Src2:Pointer;Size,Alpha:Integer);
// Dst   = eax
// Src1  = edx
// Src2  = ecx
// Size  = [ebp + 12]
// Alpha = [ebp +  8]
asm
  push ebx
  push edi
  push esi

  mov edi,edx
  mov esi,ecx

  mov ecx,[ebp+12]
  shr ecx,3
  jz  @skip

  mov ebx,[ebp+8]
  shl ebx,16
  or  ebx,[ebp+8]

  db $0F,$6E,$F3     /// movd   mm6,ebx
  db $0F,$6E,$EB     /// movd   mm5,ebx
  db $0F,$73,$F5,$20 /// psllq  mm5,32
  db $0F,$EB,$F5     /// por    mm6,mm5 // mm6 = 00aa00aa00aa00aa
  db $0F,$EF,$FF     /// pxor   mm7,mm7

  @quads:
    db $0F,$6F,$07     /// movq       mm0,[edi] // src1
    db $0F,$6F,$0E     /// movq       mm1,[esi] // src2
    db $0F,$6F,$D0     /// movq       mm2,mm0   // second copy of src1
    db $0F,$6F,$D9     /// movq       mm3,mm1   // second copy of src2
    db $0F,$6F,$E1     /// movq       mm4,mm1   // third copy of src2
    db $0F,$60,$C7     /// punpcklbw  mm0,mm7   // mm0 = unpacked low half of src1
    db $0F,$60,$CF     /// punpcklbw  mm1,mm7   // mm1 = unpacked low half of src2
    db $0F,$68,$D7     /// punpckhbw  mm2,mm7   // mm2 = unpacked high half of src1
    db $0F,$68,$DF     /// punpckhbw  mm3,mm7   // mm3 = unpacked high half of src2
    db $0F,$F9,$C1     /// psubw      mm0,mm1   // mm0 = low half of (src1-src2)
    db $0F,$F9,$D3     /// psubw      mm2,mm3   // mm2 = high half of (src1-src2)
    db $0F,$D5,$C6     /// pmullw     mm0,mm6   // low (src1-src2)*alpha
    db $0F,$D5,$D6     /// pmullw     mm2,mm6   // high (src1-src2)*alpha
    db $0F,$71,$D0,$08 /// psrlw      mm0,8     // low (src1-src2)*alpha / 256
    db $0F,$71,$D2,$08 /// psrlw      mm2,8     // high (src1-src2)*alpha / 256
    db $0F,$67,$C2     /// packuswb   mm0,mm2   // combine with unsigned saturation
    db $0F,$FC,$C4     /// paddb      mm0,mm4   // (src1-src2)*alpha / 256 + src2
    db $0F,$7F,$00     /// movq       [eax],mm0 // store the result
    add eax,8
    add edi,8
    add esi,8
    dec ecx
  jnz @quads

  @skip:
  mov ecx,[ebp+12]
  and ecx,111b
  jz  @exit
  add ecx,eax
  mov [ebp+12],ecx

  @bytes:
    mov  bl,[edi]
    mov  cl,[esi]
    and  ebx,$FF
    and  ecx,$FF
    sub  ebx,ecx
    imul ebx,[ebp+8]
    shr  ebx,8
    add  ebx,ecx
    mov  [eax],bl
    inc  eax
    inc  edi
    inc  esi
    cmp  eax,[ebp+12]
  jne @bytes

  @exit:
  pop esi
  pop edi
  pop ebx

  db $0F,$77 // emms
end;

procedure AlphaBlend8(Dst,Src1,Src2:TFastDIB;Alpha:Integer);
type
  TBlendMem8Proc = procedure(Dst,Src1,Src2:Pointer;Size,Alpha:Integer);
var
  i: Integer;
  BlendMem: TBlendMem8Proc;
begin
  if cfMMX in CPUInfo.Features then
    BlendMem:=BlendMem8MMX else BlendMem:=BlendMem8;

  for i:=0 to Dst.AbsHeight-1 do
    BlendMem(Dst.Scanlines[i],Src1.Scanlines[i],Src2.Scanlines[i],Dst.BWidth-Dst.Gap,Alpha);
end;

procedure AlphaBlend(Dst,Src1,Src2:TFastDIB;Alpha:Integer);
begin
  case Dst.Bpp of
    8,24,32: AlphaBlend8(Dst,Src1,Src2,Alpha);
    16: AlphaBlend16(Dst,Src1,Src2,Alpha);
  end;
end;

end.
